import ast

from kirin import ir, types, lowering
from kirin.dialects.py.unpack import unpacking

from .stmts import For, Yield, IfElse
from ._dialect import dialect


@dialect.register
class Lowering(lowering.FromPythonAST):

    @staticmethod
    def _frame_or_any_parent_has_def(frame, name) -> ir.SSAValue | None:
        # NOTE: this recursively checks all parents of the current frame for the
        # def. Required for nested if statements that e.g. assign to variables
        # defined in outer scope
        if frame is None:
            return None

        if name in frame.defs:
            value = frame.get(name)
            if value is None:
                raise lowering.BuildError(f"expected value for {name}")
            return value

        return Lowering._frame_or_any_parent_has_def(frame.parent, name)

    def lower_If(self, state: lowering.State, node: ast.If) -> lowering.Result:
        cond = state.lower(node.test).expect_one()
        frame = state.current_frame

        with state.frame(node.body, finalize_next=False) as body_frame:
            then_cond = body_frame.curr_block.args.append_from(types.Bool, cond.name)
            if cond.name:
                body_frame.defs[cond.name] = then_cond
            body_frame.exhaust()

        with state.frame(node.orelse, finalize_next=False) as else_frame:
            else_cond = else_frame.curr_block.args.append_from(types.Bool, cond.name)
            if cond.name:
                else_frame.defs[cond.name] = else_cond
            else_frame.exhaust()

        yield_names: list[str] = []
        body_yields: list[ir.SSAValue] = []
        else_yields: list[ir.SSAValue] = []
        for name in body_frame.defs.keys():
            if name in else_frame.defs:
                yield_names.append(name)
                body_yields.append(body_frame[name])
                else_yields.append(else_frame[name])
            elif (value := self._frame_or_any_parent_has_def(frame, name)) is not None:
                yield_names.append(name)
                body_yields.append(body_frame[name])
                else_yields.append(value)

        if not (
            body_frame.curr_block.last_stmt
            and body_frame.curr_block.last_stmt.has_trait(ir.IsTerminator)
        ):
            body_frame.push(Yield(*body_yields))

        if not (
            else_frame.curr_block.last_stmt
            and else_frame.curr_block.last_stmt.has_trait(ir.IsTerminator)
        ):
            else_frame.push(Yield(*else_yields))

        stmt = IfElse(
            cond,
            then_body=body_frame.curr_region,
            else_body=else_frame.curr_region,
        )
        for result, name, body, else_ in zip(
            stmt.results, yield_names, body_yields, else_yields
        ):
            result.name = name
            result.type = body.type.join(else_.type)
            frame.defs[name] = result
        state.current_frame.push(stmt)

    def lower_For(self, state: lowering.State, node: ast.For) -> lowering.Result:
        iter_ = state.lower(node.iter).expect_one()

        yields: list[str] = []
        parent_frame = state.current_frame

        def new_block_arg_if_inside_loop(frame: lowering.Frame, capture: ir.SSAValue):
            if not capture.name:
                raise lowering.BuildError("unexpected loop variable captured")
            yields.append(capture.name)
            return frame.curr_block.args.append_from(capture.type, capture.name)

        with state.frame(
            node.body,
            capture_callback=new_block_arg_if_inside_loop,
            finalize_next=False,
        ) as body_frame:
            loop_var = body_frame.curr_block.args.append_from(types.Any)
            unpacking(state, node.target, loop_var)
            body_frame.exhaust()

            # if a variable is assigned in loop body and exist in parent frame
            # it should be captured as initializers and yielded
            for name, value in body_frame.defs.items():
                if name in parent_frame.defs:
                    yields.append(name)
                    body_frame.curr_block.args.append_from(value.type, name)

            # NOTE: this frame won't have phi nodes
            if yields and (
                body_frame.curr_block.last_stmt is None
                or not body_frame.curr_block.last_stmt.has_trait(ir.IsTerminator)
            ):
                body_frame.push(Yield(*[body_frame.defs[name] for name in yields]))  # type: ignore

        initializers: list[ir.SSAValue] = []
        for name in yields:
            value = state.current_frame.get(name)
            if value is None:
                raise lowering.BuildError(f"expected value for {name}")
            initializers.append(value)
        stmt = For(iter_, body_frame.curr_region, *initializers)

        for name, result in zip(yields, stmt.results):
            state.current_frame.defs[name] = result
            result.name = name
        state.current_frame.push(stmt)



from typing import Generic, TypeVar
from dataclasses import field, dataclass

T = TypeVar("T")


@dataclass
class SymbolTable(Generic[T]):
    names: dict[str, T] = field(default_factory=dict)
    """The table that maps names to values."""
    prefix: str = field(default="", kw_only=True)
    name_count: dict[str, int] = field(default_factory=dict, kw_only=True)
    """The count of names that have been requested."""

    def __getitem__(self, name: str) -> T:
        return self.names[name]

    def __contains__(self, name: str) -> bool:
        return name in self.names

    def __setitem__(self, name: str, value: T) -> None:
        count = self.name_count.setdefault(name, 0)
        self.name_count[name] = count + 1
        self.names[f"{self.prefix}_{name}_{count}"] = value

    def __delitem__(self, name: str) -> None:
        del self.names[name]



from typing import Generic, TypeVar
from dataclasses import field, dataclass

T = TypeVar("T")


@dataclass
class IdTable(Generic[T]):
    """A table that maps values to "human readable" unique names.
    This is used for IR printing and code generation of SSA values
    and basic blocks, or anything else required to have a unique name.

    ## Example

    ```python
    from kirin import ir
    from kirin.idtable import IdTable
    table = IdTable()
    x = ir.TestValue()
    table[x] # "%0"
    table[x] # "%0"
    y = ir.TestValue()
    table[y] # "%1"
    ```
    """

    prefix: str = "%"
    """The prefix to use for generated names."""
    table: dict[T, str] = field(default_factory=dict)
    """The table that maps values to names."""
    name_count: dict[str, int] = field(default_factory=dict)
    """The count of names that have been generated."""
    next_id: int = 0
    """The next ID to use for generating names."""
    prefix_if_none: str = ""
    """An alternate prefix to use when the name is None."""

    def add(self, value: T) -> str:
        """Add a value to the table and return the name."""
        id = self.next_id
        if (value_name := getattr(value, "name", None)) is not None:
            curr_ind = self.name_count.get(value_name, 0)
            suffix = f"_{curr_ind}" if curr_ind != 0 else ""
            self.name_count[value_name] = curr_ind + 1
            name = self.prefix + value_name + suffix
            self.table[value] = name
        else:
            name = f"{self.prefix}{self.prefix_if_none}{id}"
            self.next_id += 1
            self.table[value] = name
        return name

    def __getitem__(self, value: T) -> str:
        if value in self.table:
            return self.table[value]
        else:
            return self.add(value)



# This file is generated by gen.py
import numpy as np

from kirin.prelude import basic_no_opt


@basic_no_opt
def numpy_passing(x: np.ndarray):
    return x


numpy_passing.code.print()


def test_passing():
    truth = np.arange(3)

    assert numpy_passing(truth) is truth



from __future__ import annotations

import ast
import builtins
from typing import Any
from dataclasses import dataclass

from kirin.source import SourceInfo
from kirin.lowering.frame import Frame
from kirin.lowering.exception import BuildError


class GlobalEvalError(BuildError):
    """Exception raised when a global expression cannot be evaluated."""

    def __init__(self, node: ast.AST, *msgs: str, help: str | None = None):
        super().__init__(*msgs, help=help)
        self.source = SourceInfo.from_ast(node)


@dataclass
class GlobalExprEval(ast.NodeVisitor):
    frame: Frame

    def generic_visit(self, node: ast.AST) -> Any:
        if isinstance(node, ast.AST):
            raise GlobalEvalError(
                node,
                f"Cannot lower global {node.__class__.__name__} node: {ast.dump(node)}",
            )
        raise GlobalEvalError(
            node,
            f"Unexpected global `{node.__class__.__name__}` node: {repr(node)} is not an AST node",
        )

    def visit_Name(self, node: ast.Name) -> Any:
        if not isinstance(node.ctx, ast.Load):
            raise GlobalEvalError(node, "unsupported name access")

        name = node.id
        value = self.frame.globals.get(name)
        if value is not None:
            return value

        if hasattr(builtins, name):
            return getattr(builtins, name)
        else:
            raise GlobalEvalError(node, f"global {name} not found")

    def visit_Constant(self, node: ast.Constant) -> Any:
        return node.value

    def visit_Attribute(self, node: ast.Attribute) -> Any:
        if not isinstance(node.ctx, ast.Load):
            raise GlobalEvalError(node, "unsupported attribute access")

        value = self.visit(node.value)
        if hasattr(value, node.attr):
            return getattr(value, node.attr)

        raise GlobalEvalError(node, f"attribute {node.attr} not found in {value}")

    def visit_Subscript(self, node: ast.Subscript) -> Any:
        value = self.visit(node.value)

        if isinstance(value, type):
            if hasattr(value, "__class_getitem__"):
                return value.__class_getitem__(self.visit(node.slice))
            else:
                raise GlobalEvalError(
                    node, f"class {value} does not support type parameters"
                )

        if not hasattr(value, "__getitem__"):
            raise GlobalEvalError(
                node, f"unsupported subscript access for class {type(value)}"
            )

        return value[self.visit(node.slice)]

    def visit_Call(self, node: ast.Call) -> Any:
        func = self.visit(node.func)
        args = [self.visit(arg) for arg in node.args]
        keywords = {
            kw.arg: self.visit(kw) for kw in node.keywords if kw.arg is not None
        }
        if not callable(func):
            raise GlobalEvalError(node, f"global object {func} is not callable")

        try:
            return func(*args, **keywords)
        except TypeError as e:
            raise GlobalEvalError(
                node, f"TypeError in global call: {e} for {func}({args}, {keywords})"
            ) from e
        except Exception as e:
            raise GlobalEvalError(
                node, f"Exception in global call: {e} for {func}({args}, {keywords})"
            ) from e

    def visit_Tuple(self, node: ast.Tuple) -> Any:
        return tuple(self.visit(elt) for elt in node.elts)

    def visit_List(self, node: ast.List) -> Any:
        return [self.visit(elt) for elt in node.elts]



from dataclasses import dataclass

from kirin.rewrite import Walk, Chain, Fixpoint
from kirin.analysis import const
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.rewrite.dce import DeadCodeElimination
from kirin.rewrite.fold import ConstantFold
from kirin.ir.nodes.base import IRNode
from kirin.rewrite.inline import Inline
from kirin.rewrite.getitem import InlineGetItem
from kirin.rewrite.getfield import InlineGetField
from kirin.rewrite.compactify import CFGCompactify
from kirin.rewrite.wrap_const import WrapConst
from kirin.rewrite.call2invoke import Call2Invoke
from kirin.rewrite.type_assert import InlineTypeAssert


@dataclass
class Fold(RewriteRule):
    rule: RewriteRule

    def __init__(self, frame: const.Frame):
        rule = Fixpoint(
            Chain(
                Walk(WrapConst(frame)),
                Walk(Inline(lambda _: True)),
                Walk(ConstantFold()),
                Walk(Call2Invoke()),
                Fixpoint(
                    Walk(
                        Chain(
                            InlineTypeAssert(),
                            InlineGetItem(),
                            InlineGetField(),
                            DeadCodeElimination(),
                        )
                    )
                ),
                Walk(CFGCompactify()),
            )
        )
        self.rule = rule

    def rewrite(self, node: IRNode) -> RewriteResult:
        return self.rule.rewrite(node)



from kirin.prelude import basic_no_opt
from kirin.dialects import debug


def test_debug_printing():
    @basic_no_opt.add(debug)
    def test_if_inside_for() -> int:
        count = 0
        for i in range(5):
            count = count + 1
            something_else = count + 2
            debug.info("current count before", count, something_else)
            if True:
                count = count + 100
                debug.info("inside the ifelse", count, something_else)
            else:
                count = count + 300
        return count

    test_if_inside_for()



from typing import IO, TypeVar

from kirin import ir
from kirin.ir.attrs.types import PyClass
from kirin.ir.nodes.block import Block

from .str import EmitStr, EmitStrFrame

IO_t = TypeVar("IO_t", bound=IO)


class EmitJulia(EmitStr[IO_t]):
    keys = ["emit.julia"]

    PYTYPE_MAP = {
        int: "Int",
        float: "Real",
        str: "String",
        bool: "Bool",
        type(None): "Nothing",
        dict: "Dict",
        list: "Vector",
        tuple: "Tuple",
    }

    def emit_block_begin(self, frame: EmitStrFrame, block: Block) -> None:
        block_id = self.block_id[block]
        frame.block_ref[block] = block_id
        self.newline(frame)
        self.write(f"@label {block_id};")

    def emit_type_PyClass(self, attr: PyClass) -> str:
        return self.PYTYPE_MAP.get(attr.typ, "Any")

    def write_assign(self, frame: EmitStrFrame, result: ir.SSAValue, *args):
        result_sym = self.ssa_id[result]
        frame.set(result, result_sym)
        self.writeln(frame, result_sym, " = ", *args)
        return result_sym

    def emit_binaryop(
        self,
        frame: EmitStrFrame,
        sym: str,
        lhs: ir.SSAValue,
        rhs: ir.SSAValue,
        result: ir.ResultValue,
    ):
        return (
            self.write_assign(
                frame,
                result,
                f"{frame.get(lhs)} {sym} {frame.get(rhs)}",
            ),
        )

    def emit_type_PyAttr(self, attr: ir.PyAttr) -> str:
        if isinstance(attr.data, (int, float)):
            return repr(attr.data)
        elif isinstance(attr.data, str):
            return f'"{attr.data}"'
        else:
            raise ValueError(f"unsupported type {type(attr.data)}")



from kirin import ir, types
from kirin.prelude import basic
from kirin.analysis import const


@basic(typeinfer=True, fold=True)
def unstable(x: int):  # type: ignore
    y = x + 1
    if y > 10:
        z = y
    else:
        z = y + 1.2
    return z


def test_apply_type():
    def stmt_at(block_id, stmt_id: int) -> ir.Statement:
        return unstable.callable_region.blocks[block_id].stmts.at(stmt_id)  # type: ignore

    assert stmt_at(0, 0).results.types == [types.Int]
    assert stmt_at(0, 0).results[0].hints.get("const") == const.Value(1)
    assert stmt_at(0, 1).results.types == [types.Int]
    assert stmt_at(0, 2).results.types == [types.Int]
    assert stmt_at(0, 2).results[0].hints.get("const") == const.Value(10)
    assert stmt_at(0, 3).results.types == [types.Bool]

    assert stmt_at(1, 0).results.types == [types.Int]
    assert stmt_at(2, 0).results.types == [types.Float]
    assert stmt_at(2, 0).results[0].hints.get("const") == const.Value(1.2)
    assert stmt_at(2, 1).results.types == [types.Float]

    stmt = stmt_at(3, 0)
    assert stmt.args[0].type == (types.Int | types.Float)



# This file is generated by gen.py
import math

from kirin.interp import Frame, MethodTable, impl
from kirin.dialects.math import stmts
from kirin.dialects.math.dialect import dialect


@dialect.register
class MathMethodTable(MethodTable):

    @impl(stmts.acos)
    def acos(self, interp, frame: Frame, stmt: stmts.acos):
        values = frame.get_values(stmt.args)
        return (math.acos(values[0]),)

    @impl(stmts.asin)
    def asin(self, interp, frame: Frame, stmt: stmts.asin):
        values = frame.get_values(stmt.args)
        return (math.asin(values[0]),)

    @impl(stmts.asinh)
    def asinh(self, interp, frame: Frame, stmt: stmts.asinh):
        values = frame.get_values(stmt.args)
        return (math.asinh(values[0]),)

    @impl(stmts.atan)
    def atan(self, interp, frame: Frame, stmt: stmts.atan):
        values = frame.get_values(stmt.args)
        return (math.atan(values[0]),)

    @impl(stmts.atan2)
    def atan2(self, interp, frame: Frame, stmt: stmts.atan2):
        values = frame.get_values(stmt.args)
        return (math.atan2(values[0], values[1]),)

    @impl(stmts.atanh)
    def atanh(self, interp, frame: Frame, stmt: stmts.atanh):
        values = frame.get_values(stmt.args)
        return (math.atanh(values[0]),)

    @impl(stmts.ceil)
    def ceil(self, interp, frame: Frame, stmt: stmts.ceil):
        values = frame.get_values(stmt.args)
        return (math.ceil(values[0]),)

    @impl(stmts.copysign)
    def copysign(self, interp, frame: Frame, stmt: stmts.copysign):
        values = frame.get_values(stmt.args)
        return (math.copysign(values[0], values[1]),)

    @impl(stmts.cos)
    def cos(self, interp, frame: Frame, stmt: stmts.cos):
        values = frame.get_values(stmt.args)
        return (math.cos(values[0]),)

    @impl(stmts.cosh)
    def cosh(self, interp, frame: Frame, stmt: stmts.cosh):
        values = frame.get_values(stmt.args)
        return (math.cosh(values[0]),)

    @impl(stmts.degrees)
    def degrees(self, interp, frame: Frame, stmt: stmts.degrees):
        values = frame.get_values(stmt.args)
        return (math.degrees(values[0]),)

    @impl(stmts.erf)
    def erf(self, interp, frame: Frame, stmt: stmts.erf):
        values = frame.get_values(stmt.args)
        return (math.erf(values[0]),)

    @impl(stmts.erfc)
    def erfc(self, interp, frame: Frame, stmt: stmts.erfc):
        values = frame.get_values(stmt.args)
        return (math.erfc(values[0]),)

    @impl(stmts.exp)
    def exp(self, interp, frame: Frame, stmt: stmts.exp):
        values = frame.get_values(stmt.args)
        return (math.exp(values[0]),)

    @impl(stmts.expm1)
    def expm1(self, interp, frame: Frame, stmt: stmts.expm1):
        values = frame.get_values(stmt.args)
        return (math.expm1(values[0]),)

    @impl(stmts.fabs)
    def fabs(self, interp, frame: Frame, stmt: stmts.fabs):
        values = frame.get_values(stmt.args)
        return (math.fabs(values[0]),)

    @impl(stmts.floor)
    def floor(self, interp, frame: Frame, stmt: stmts.floor):
        values = frame.get_values(stmt.args)
        return (math.floor(values[0]),)

    @impl(stmts.fmod)
    def fmod(self, interp, frame: Frame, stmt: stmts.fmod):
        values = frame.get_values(stmt.args)
        return (math.fmod(values[0], values[1]),)

    @impl(stmts.gamma)
    def gamma(self, interp, frame: Frame, stmt: stmts.gamma):
        values = frame.get_values(stmt.args)
        return (math.gamma(values[0]),)

    @impl(stmts.isfinite)
    def isfinite(self, interp, frame: Frame, stmt: stmts.isfinite):
        values = frame.get_values(stmt.args)
        return (math.isfinite(values[0]),)

    @impl(stmts.isinf)
    def isinf(self, interp, frame: Frame, stmt: stmts.isinf):
        values = frame.get_values(stmt.args)
        return (math.isinf(values[0]),)

    @impl(stmts.isnan)
    def isnan(self, interp, frame: Frame, stmt: stmts.isnan):
        values = frame.get_values(stmt.args)
        return (math.isnan(values[0]),)

    @impl(stmts.lgamma)
    def lgamma(self, interp, frame: Frame, stmt: stmts.lgamma):
        values = frame.get_values(stmt.args)
        return (math.lgamma(values[0]),)

    @impl(stmts.log10)
    def log10(self, interp, frame: Frame, stmt: stmts.log10):
        values = frame.get_values(stmt.args)
        return (math.log10(values[0]),)

    @impl(stmts.log1p)
    def log1p(self, interp, frame: Frame, stmt: stmts.log1p):
        values = frame.get_values(stmt.args)
        return (math.log1p(values[0]),)

    @impl(stmts.log2)
    def log2(self, interp, frame: Frame, stmt: stmts.log2):
        values = frame.get_values(stmt.args)
        return (math.log2(values[0]),)

    @impl(stmts.pow)
    def pow(self, interp, frame: Frame, stmt: stmts.pow):
        values = frame.get_values(stmt.args)
        return (math.pow(values[0], values[1]),)

    @impl(stmts.radians)
    def radians(self, interp, frame: Frame, stmt: stmts.radians):
        values = frame.get_values(stmt.args)
        return (math.radians(values[0]),)

    @impl(stmts.remainder)
    def remainder(self, interp, frame: Frame, stmt: stmts.remainder):
        values = frame.get_values(stmt.args)
        return (math.remainder(values[0], values[1]),)

    @impl(stmts.sin)
    def sin(self, interp, frame: Frame, stmt: stmts.sin):
        values = frame.get_values(stmt.args)
        return (math.sin(values[0]),)

    @impl(stmts.sinh)
    def sinh(self, interp, frame: Frame, stmt: stmts.sinh):
        values = frame.get_values(stmt.args)
        return (math.sinh(values[0]),)

    @impl(stmts.sqrt)
    def sqrt(self, interp, frame: Frame, stmt: stmts.sqrt):
        values = frame.get_values(stmt.args)
        return (math.sqrt(values[0]),)

    @impl(stmts.tan)
    def tan(self, interp, frame: Frame, stmt: stmts.tan):
        values = frame.get_values(stmt.args)
        return (math.tan(values[0]),)

    @impl(stmts.tanh)
    def tanh(self, interp, frame: Frame, stmt: stmts.tanh):
        values = frame.get_values(stmt.args)
        return (math.tanh(values[0]),)

    @impl(stmts.trunc)
    def trunc(self, interp, frame: Frame, stmt: stmts.trunc):
        values = frame.get_values(stmt.args)
        return (math.trunc(values[0]),)

    @impl(stmts.ulp)
    def ulp(self, interp, frame: Frame, stmt: stmts.ulp):
        values = frame.get_values(stmt.args)
        return (math.ulp(values[0]),)



from kirin import ir, types, lowering
from kirin.decl import info, statement
from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk
from kirin.dialects.py import indexing

dummy = ir.Dialect("dummy")


class RegGetItemInterface(indexing.GetItemLike["RegGetItem"]):

    def get_object(self, stmt: "RegGetItem") -> ir.SSAValue:
        return stmt.reg

    def get_index(self, stmt: "RegGetItem") -> ir.SSAValue:
        return stmt.index

    def new(
        self, stmt_type: type["RegGetItem"], obj: ir.SSAValue, index: ir.SSAValue
    ) -> "RegGetItem":
        return RegGetItem(obj, index)


class Register:
    pass


@statement(dialect=dummy)
class New(ir.Statement):
    name = "new"
    traits = frozenset({ir.Pure(), lowering.FromPythonCall()})
    result: ir.ResultValue = info.result(types.PyClass(Register))


@statement(dialect=dummy)
class RegGetItem(ir.Statement):
    name = "reg.get"
    traits = frozenset({ir.Pure(), RegGetItemInterface()})
    reg: ir.SSAValue = info.argument(types.PyClass(Register))
    index: ir.SSAValue = info.argument(types.Int)
    result: ir.ResultValue = info.result(types.Int)


@basic_no_opt.add(dummy)
def main():
    reg = New()
    return reg[0]  # type: ignore


def test_rewrite_getitem():
    rule = Walk(indexing.RewriteGetItem(RegGetItem, types.PyClass(Register)))

    stmt: ir.Statement = main.code.body.blocks[0].stmts.at(-2)  # type: ignore
    assert isinstance(stmt, indexing.GetItem)
    rule.rewrite(main.code)
    stmt: ir.Statement = main.code.body.blocks[0].stmts.at(-2)  # type: ignore
    assert isinstance(stmt, RegGetItem)



from kirin import types, interp
from kirin.dialects.eltype import ElType
from kirin.dialects.py.binop import Add
from kirin.dialects.py.indexing import GetItem

from ._dialect import dialect


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(ElType, types.PyClass(list))
    def eltype_list(self, interp, frame: interp.Frame, stmt: ElType):
        list_type = frame.get(stmt.container)
        if isinstance(list_type, types.Generic):
            return (list_type.vars[0],)
        else:
            return (types.Any,)

    @interp.impl(Add, types.PyClass(list), types.PyClass(list))
    def add(self, interp, frame: interp.Frame, stmt: Add):
        lhs_type = frame.get(stmt.lhs)
        rhs_type = frame.get(stmt.rhs)
        if isinstance(lhs_type, types.Generic):
            lhs_elem_type = lhs_type.vars[0]
        else:
            lhs_elem_type = types.Any

        if isinstance(rhs_type, types.Generic):
            rhs_elem_type = rhs_type.vars[0]
        else:
            rhs_elem_type = types.Any

        return (types.List[lhs_elem_type.join(rhs_elem_type)],)

    @interp.impl(GetItem, types.PyClass(list), types.Int)
    def getitem_list_int(
        self, interp, frame: interp.Frame[types.TypeAttribute], stmt: GetItem
    ):
        obj_type = frame.get(stmt.obj)
        if isinstance(obj_type, types.Generic):
            return (obj_type.vars[0],)
        else:
            return (types.Any,)

    @interp.impl(GetItem, types.PyClass(list), types.PyClass(slice))
    def getitem_list_slice(
        self, interp, frame: interp.Frame[types.TypeAttribute], stmt: GetItem
    ):
        obj_type = frame.get(stmt.obj)
        if isinstance(obj_type, types.Generic):
            return (types.List[obj_type.vars[0]],)
        else:
            return (types.Any,)



import pytest

from kirin.prelude import basic_no_opt
from kirin.dialects.py import builtin


@basic_no_opt
def abs_kernel(x):
    return builtin.Abs(value=x)


@pytest.mark.parametrize("x", [-1, 2, 3.0, -3.2])
def test_abs(x):

    abs_x = abs_kernel(x)

    assert isinstance(abs_x, type(abs(x)))
    assert abs_x == abs(x)



from dataclasses import dataclass

from stmts import Eat, Nap, NewFood, RandomBranch

from kirin import ir
from kirin.dialects import cf
from kirin.rewrite.abc import RewriteRule, RewriteResult


@dataclass
class RandomWalkBranch(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, cf.ConditionalBranch):
            return RewriteResult()
        node.replace_by(
            RandomBranch(
                cond=node.cond,
                then_arguments=node.then_arguments,
                then_successor=node.then_successor,
                else_arguments=node.else_arguments,
                else_successor=node.else_successor,
            )
        )
        return RewriteResult(has_done_something=True)


@dataclass
class NewFoodAndNap(RewriteRule):
    # sometimes someone is hungry and needs a nap
    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, Eat):
            return RewriteResult()

        # 1. create new stmts:
        new_food_stmt = NewFood(type="burger")
        nap_stmt = Nap()

        # 2. put them in the ir
        new_food_stmt.insert_after(node)
        nap_stmt.insert_after(new_food_stmt)

        return RewriteResult(has_done_something=True)



from typing import final
from dataclasses import dataclass

from kirin.lattice import (
    SingletonMeta,
    BoundedLattice,
    IsSubsetEqMixin,
    SimpleJoinMixin,
    SimpleMeetMixin,
)


@dataclass
class Item(
    IsSubsetEqMixin["Item"],
    SimpleJoinMixin["Item"],
    SimpleMeetMixin["Item"],
    BoundedLattice["Item"],
):

    @classmethod
    def top(cls) -> "Item":
        return AnyItem()

    @classmethod
    def bottom(cls) -> "Item":
        return NotItem()


@final
@dataclass
class NotItem(Item, metaclass=SingletonMeta):
    """The bottom of the lattice.

    Since the element is the same without any field,
    we can use the SingletonMeta to make it a singleton by inherit the metaclass

    """

    def is_subseteq(self, other: Item) -> bool:
        return True


@final
@dataclass
class AnyItem(Item, metaclass=SingletonMeta):
    """The top of the lattice.

    Since the element is the same without any field,
    we can use the SingletonMeta to make it a singleton by inherit the metaclass

    """

    def is_subseteq(self, other: Item) -> bool:
        return isinstance(other, AnyItem)


@final
@dataclass
class ItemServing(Item):
    count: Item
    type: str

    def is_subseteq(self, other: Item) -> bool:
        return (
            isinstance(other, ItemServing)
            and self.count == other.count
            and self.type == other.type
        )


@final
@dataclass
class AtLeastXItem(Item):
    data: int

    def is_subseteq(self, other: Item) -> bool:
        return isinstance(other, AtLeastXItem) and self.data == other.data


@final
@dataclass
class ConstIntItem(Item):
    data: int

    def is_subseteq(self, other: Item) -> bool:
        return isinstance(other, ConstIntItem) and self.data == other.data


@final
@dataclass
class ItemFood(Item):
    type: str

    def is_subseteq(self, other: Item) -> bool:
        return isinstance(other, ItemFood) and self.type == other.type



from kirin import ir, types, lowering
from kirin.prelude import basic_no_opt
from kirin.rewrite import Walk, Chain, Fixpoint, compactify
from kirin.analysis import CFG
from kirin.dialects import cf, py, func


def test_duplicated_branch():
    code = func.Function(
        sym_name="duplicated_branch",
        signature=func.Signature((), types.NoneType),
        body=ir.Region(),
    )
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())

    true = py.Constant(True)
    code.body.blocks[0].stmts.append(true)
    code.body.blocks[0].stmts.append(
        cf.ConditionalBranch(
            true.result,
            (),
            (),
            then_successor=code.body.blocks[1],
            else_successor=code.body.blocks[1],
        )
    )
    code.body.blocks[1].stmts.append(cf.Branch((), successor=code.body.blocks[2]))
    code.body.blocks[2].stmts.append(cf.Branch((), successor=code.body.blocks[3]))
    code.body.blocks[3].stmts.append(cf.Branch((), successor=code.body.blocks[4]))
    none = func.ConstantNone()
    code.body.blocks[4].stmts.append(none)
    code.body.blocks[4].stmts.append(func.Return(none.result))
    Fixpoint(
        Walk(Chain(compactify.DuplicatedBranch(), compactify.CFGEdge(CFG(code.body))))
    ).rewrite(code)

    target = func.Function(
        sym_name="duplicated_branch",
        signature=func.Signature((), types.NoneType),
        body=ir.Region(),
    )
    target.body.blocks.append(ir.Block())
    bb0 = target.body.blocks[0]
    true = py.Constant(True)
    none = func.ConstantNone()
    bb0.stmts.append(true)
    bb0.stmts.append(none)
    bb0.stmts.append(func.Return(none.result))

    assert code.is_equal(target)


def test_cfg_skip_block():
    code = func.Function(
        sym_name="cfg_double_branch",
        signature=func.Signature((), types.NoneType),
        body=ir.Region(),
    )
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())
    code.body.blocks.append(ir.Block())
    code.body.blocks[0].args.append_from(types.Any, "self")
    cond = code.body.blocks[0].args.append_from(types.Any, "cond")
    a = py.Constant(1)
    b = py.Constant(2)
    c = py.Constant(3)

    code.body.blocks[0].stmts.append(a)
    code.body.blocks[0].stmts.append(b)
    code.body.blocks[0].stmts.append(c)
    code.body.blocks[0].stmts.append(
        cf.ConditionalBranch(
            cond,
            (a.result,),
            (b.result, c.result),
            then_successor=code.body.blocks[1],
            else_successor=code.body.blocks[2],
        )
    )

    bb1_a = code.body.blocks[1].args.append_from(types.Any, "a")
    code.body.blocks[1].stmts.append(cf.Branch((bb1_a,), successor=code.body.blocks[3]))

    bb2_b = code.body.blocks[2].args.append_from(types.Any, "b")
    code.body.blocks[2].args.append_from(types.Any, "c")
    code.body.blocks[2].stmts.append(cf.Branch((bb2_b,), successor=code.body.blocks[3]))

    code.body.blocks[3].args.append_from(types.Any, "x")
    none = func.ConstantNone()
    code.body.blocks[3].stmts.append(none)
    code.body.blocks[3].stmts.append(func.Return(none.result))

    Fixpoint(Walk(compactify.SkipBlock(CFG(code.body)))).rewrite(code)
    Fixpoint(Walk(compactify.CFGEdge(CFG(code.body)))).rewrite(code)
    Fixpoint(Walk(compactify.DeadBlock(CFG(code.body)))).rewrite(code)

    target = func.Function(
        sym_name="cfg_double_branch",
        signature=func.Signature((), types.NoneType),
        body=ir.Region(),
    )
    target.body.blocks.append(ir.Block())
    target.body.blocks.append(ir.Block())

    bb0 = target.body.blocks[0]
    bb1 = target.body.blocks[1]
    bb0.args.append_from(types.Any, "self")
    cond = bb0.args.append_from(types.Any, "cond")
    a = py.Constant(1)
    b = py.Constant(2)
    c = py.Constant(3)
    bb0.stmts.extend([a, b, c])
    bb0.stmts.append(
        cf.ConditionalBranch(
            cond, (a.result,), (b.result,), then_successor=bb1, else_successor=bb1
        )
    )

    bb1.args.append_from(types.Any, "x")
    none = func.ConstantNone()
    bb1.stmts.extend([none, func.Return(none.result)])
    code.print()
    target.print()
    assert code.is_equal(target)


def test_cfg_pass_around():
    def main():
        if True:
            pass
        else:
            pass

    lower = lowering.Python(basic_no_opt)
    code = lower.python_function(main, compactify=False)
    assert isinstance(code, func.Function)

    cfg = CFG(code.body)
    compactify.DeadBlock(cfg).rewrite(code.body)
    compactify.CFGEdge(cfg).rewrite(code.body)
    assert cfg.successors == CFG(code.body).successors
    compactify.CFGEdge(cfg).rewrite(code.body)
    assert cfg.successors == CFG(code.body).successors
    compactify.SkipBlock(cfg).rewrite(code.body)
    compactify.DeadBlock(cfg).rewrite(code.body)
    assert cfg.successors == CFG(code.body).successors
    Walk(compactify.DuplicatedBranch()).rewrite(code.body)
    assert cfg.successors == CFG(code.body).successors
    compactify.CFGEdge(cfg).rewrite(code.body)
    assert cfg.successors == CFG(code.body).successors



from dataclasses import dataclass

from kirin import ir
from kirin.rewrite.abc import RewriteRule, RewriteResult
from kirin.dialects.py.assign import Alias


@dataclass
class InlineAlias(RewriteRule):

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if not isinstance(node, Alias):
            return RewriteResult()

        node.result.replace_by(node.value)
        return RewriteResult(has_done_something=True)



from typing import IO, TypeVar

from kirin import emit
from kirin.interp import Successor, MethodTable, impl
from kirin.emit.julia import EmitJulia

from .stmts import Branch, ConditionalBranch
from .dialect import dialect

IO_t = TypeVar("IO_t", bound=IO)


@dialect.register(key="emit.julia")
class JuliaMethodTable(MethodTable):

    @impl(Branch)
    def emit_branch(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: Branch
    ):
        interp.writeln(frame, f"@goto {interp.block_id[stmt.successor]};")
        frame.worklist.append(
            Successor(stmt.successor, frame.get_values(stmt.arguments))
        )
        return ()

    @impl(ConditionalBranch)
    def emit_cbr(
        self, interp: EmitJulia[IO_t], frame: emit.EmitStrFrame, stmt: ConditionalBranch
    ):
        cond = frame.get(stmt.cond)
        interp.writeln(frame, f"if {cond}")
        frame.indent += 1
        values = frame.get_values(stmt.then_arguments)
        block_values = tuple(interp.ssa_id[x] for x in stmt.then_successor.args)
        frame.set_values(stmt.then_successor.args, block_values)
        for x, y in zip(block_values, values):
            interp.writeln(frame, f"{x} = {y};")
        interp.writeln(frame, f"@goto {interp.block_id[stmt.then_successor]};")
        frame.indent -= 1
        interp.writeln(frame, "else")
        frame.indent += 1

        values = frame.get_values(stmt.else_arguments)
        block_values = tuple(interp.ssa_id[x] for x in stmt.else_successor.args)
        frame.set_values(stmt.else_successor.args, block_values)
        for x, y in zip(block_values, values):
            interp.writeln(frame, f"{x} = {y};")
        interp.writeln(frame, f"@goto {interp.block_id[stmt.else_successor]};")
        frame.indent -= 1
        interp.writeln(frame, "end")

        frame.worklist.append(
            Successor(stmt.then_successor, frame.get_values(stmt.then_arguments))
        )
        frame.worklist.append(
            Successor(stmt.else_successor, frame.get_values(stmt.else_arguments))
        )
        return ()



from kirin import ir, types
from kirin.prelude import basic, structural
from kirin.rewrite import Walk
from kirin.dialects import cf, py, func, ilist
from kirin.dialects.scf import scf2cf


def test_scf2cf_if_1():

    @structural(typeinfer=True)
    def test(b: bool):
        if b:
            b = False
        else:
            b = not b

        return b

    rule = Walk(scf2cf.ScfToCfRule())
    rule.rewrite(test.code)
    test = test.similar(basic)

    excpected_callable_region = ir.Region(
        [
            branch_block := ir.Block(),
            then_block := ir.Block(),
            else_block := ir.Block(),
            join_block := ir.Block(),
        ]
    )

    branch_block.args.append_from(types.MethodType, "self")
    b = branch_block.args.append_from(types.Bool, "b")
    branch_block.stmts.append(
        cf.ConditionalBranch(
            cond=b,
            then_arguments=(b,),
            then_successor=then_block,
            else_arguments=(b,),
            else_successor=else_block,
        )
    )

    then_block.args.append_from(types.Bool, "b")
    then_block.stmts.append(stmt := py.Constant(value=False))
    then_block.stmts.append(
        cf.Branch(
            arguments=(stmt.result,),
            successor=join_block,
        )
    )

    b = else_block.args.append_from(types.Bool)
    else_block.stmts.append(stmt := py.unary.Not(b))
    else_block.stmts.append(
        cf.Branch(
            arguments=(stmt.result,),
            successor=join_block,
        )
    )
    ret = join_block.args.append_from(types.Bool)
    join_block.stmts.append(func.Return(ret))

    expected_code = func.Function(
        sym_name="test",
        signature=func.Signature(
            output=types.Bool,
            inputs=(types.Bool,),
        ),
        body=excpected_callable_region,
    )

    expected_test = ir.Method(
        dialects=basic,
        code=expected_code,
        mod=None,
        py_func=None,
        sym_name="expected_test",
        arg_names=[],
    )

    if basic.run_pass is not None:
        basic.run_pass(expected_test, typeinfer=True)
        basic.run_pass(test, typeinfer=True)

    assert expected_test.callable_region.is_structurally_equal(test.callable_region)


def test_scf2cf_for_1():

    @structural(typeinfer=True, fold=False)
    def test():
        j = 0
        for i in range(10):
            j = j + 1

        return j

    rule = Walk(scf2cf.ScfToCfRule())
    rule.rewrite(test.code)
    test = test.similar(basic)

    expected_callable_region = ir.Region(
        [
            entry_block := ir.Block(),
            body_block := ir.Block(),
            exit_block := ir.Block(),
        ]
    )

    entry_block.args.append_from(types.MethodType, "self")
    entry_block.stmts.append(j_start := py.Constant(value=0))
    j_start.result.name = "j"
    entry_block.stmts.append(iter_start := py.Constant(value=0))
    entry_block.stmts.append(iter_end := py.Constant(value=10))
    entry_block.stmts.append(iter_step := py.Constant(value=1))
    entry_block.stmts.append(
        range_stmt := ilist.stmts.Range(
            start=iter_start.result,
            stop=iter_end.result,
            step=iter_step.result,
        )
    )
    range_stmt.result.type = ilist.IListType[types.Int, types.Literal(10)]
    entry_block.stmts.append(iterable_stmt := py.iterable.Iter(range_stmt.result))
    entry_block.stmts.append(none_stmt := func.ConstantNone())
    entry_block.stmts.append(
        first_iter := py.iterable.Next(iterable_stmt.expect_one_result())
    )
    entry_block.stmts.append(
        loop_cmp := py.cmp.Is(first_iter.expect_one_result(), none_stmt.result)
    )
    entry_block.stmts.append(
        cf.ConditionalBranch(
            cond=loop_cmp.result,
            then_arguments=(j_start.result, j_start.result),
            then_successor=exit_block,
            else_arguments=(
                first_iter.expect_one_result(),
                j_start.result,
                j_start.result,
            ),
            else_successor=body_block,
        )
    )

    body_block.args.append_from(types.Int, "i")
    body_block.args.append_from(types.Int, "j")
    body_block.args.append_from(types.Int, "j")

    body_block.stmts.append(one_stmt := py.Constant(value=1))
    body_block.stmts.append(
        j_add := py.binop.Add(
            lhs=body_block.args[1],
            rhs=one_stmt.result,
        )
    )
    j_add.result.name = "j"
    j_add.result.type = types.Int
    body_block.stmts.append(
        next_iter := py.iterable.Next(iterable_stmt.expect_one_result())
    )
    body_block.stmts.append(
        loop_cmp := py.cmp.Is(next_iter.expect_one_result(), none_stmt.result)
    )
    body_block.stmts.append(
        cf.ConditionalBranch(
            cond=loop_cmp.result,
            then_arguments=(j_add.result, j_add.result),
            then_successor=exit_block,
            else_arguments=(next_iter.expect_one_result(), j_add.result, j_add.result),
            else_successor=body_block,
        )
    )

    exit_block.args.append_from(types.Int, "j")
    exit_block.args.append_from(types.Int, "j")
    exit_block.stmts.append(func.Return(exit_block.args[1]))

    expected_code = func.Function(
        sym_name="test",
        signature=func.Signature(
            output=types.Literal(10),
            inputs=(),
        ),
        body=expected_callable_region,
    )

    expected_test = ir.Method(
        dialects=basic,
        code=expected_code,
        mod=None,
        py_func=None,
        sym_name="expected_test",
        arg_names=[],
    )

    test.print()
    expected_test.print()

    if basic.run_pass is not None:
        basic.run_pass(test, typeinfer=True, fold=False)
        basic.run_pass(expected_test, typeinfer=True, fold=False)

    assert expected_test.callable_region.is_structurally_equal(test.callable_region)



from dataclasses import dataclass

from kirin.ir.method import Method
from kirin.rewrite.abc import RewriteResult

from .abc import Pass


@dataclass
class PostInference(Pass):

    def unsafe_run(self, mt: Method) -> RewriteResult:
        result = RewriteResult()
        for dialect in self.dialects:
            for rule in dialect.rules.inference:
                result = rule.rewrite(mt.code).join(result)
        return result



"""
Immutable list dialect for Python.

This dialect provides a simple, immutable list dialect similar
to Python's built-in list type.
"""

from . import (
    interp as interp,
    rewrite as rewrite,
    lowering as lowering,
    constprop as constprop,
    typeinfer as typeinfer,
)
from .stmts import (
    Map as Map,
    New as New,
    Push as Push,
    Scan as Scan,
    Foldl as Foldl,
    Foldr as Foldr,
    ForEach as ForEach,
    IListType as IListType,
)
from .passes import IListDesugar as IListDesugar
from .runtime import IList as IList
from ._dialect import dialect as dialect
from ._wrapper import (  # careful this is not the builtin range
    map as map,
    scan as scan,
    foldl as foldl,
    foldr as foldr,
    range as range,
    for_each as for_each,
)



from dataclasses import dataclass

import pytest

from kirin import interp
from kirin.lattice import EmptyLattice
from kirin.prelude import basic
from kirin.dialects import py
from kirin.ir.method import Method
from kirin.ir.nodes.stmt import Statement
from kirin.analysis.forward import Forward, ForwardFrame


@dataclass(init=False)
class DummyInterpreter(Forward[EmptyLattice]):
    keys = ["test_interp"]
    lattice = EmptyLattice

    def run_method(self, method: Method, args: tuple[EmptyLattice, ...]):
        return self.run_callable(method.code, (EmptyLattice(),) + args)

    def eval_stmt_fallback(
        self, frame: ForwardFrame[EmptyLattice], stmt: Statement
    ) -> tuple[EmptyLattice, ...] | interp.SpecialValue[EmptyLattice]:
        ret = super().eval_stmt_fallback(frame, stmt)
        print("fallback: ", ret)
        return ret


@py.tuple.dialect.register(key="test_interp")
class DialectMethodTable(interp.MethodTable):

    @interp.impl(py.tuple.New)
    def new_tuple(self, interp: DummyInterpreter, frame, stmt: py.tuple.New):
        return (EmptyLattice(),)


@basic
def main(x):
    return 1


def test_interp():
    interp_ = DummyInterpreter(basic)
    with pytest.raises(interp.InterpreterError):
        interp_.run(main, (EmptyLattice(),))



"math dialect, modeling functions in python's `math` stdlib"  # This file is generated by gen.py

import math as pymath

from kirin import lowering
from kirin.dialects.math.dialect import dialect as dialect

from . import stmts as stmts, interp as interp

pi = pymath.pi
e = pymath.e
tau = pymath.tau


@lowering.wraps(stmts.acos)
def acos(x: float) -> float: ...


@lowering.wraps(stmts.asin)
def asin(x: float) -> float: ...


@lowering.wraps(stmts.asinh)
def asinh(x: float) -> float: ...


@lowering.wraps(stmts.atan)
def atan(x: float) -> float: ...


@lowering.wraps(stmts.atan2)
def atan2(y: float, x: float) -> float: ...


@lowering.wraps(stmts.atanh)
def atanh(x: float) -> float: ...


@lowering.wraps(stmts.ceil)
def ceil(x: float) -> float: ...


@lowering.wraps(stmts.copysign)
def copysign(x: float, y: float) -> float: ...


@lowering.wraps(stmts.cos)
def cos(x: float) -> float: ...


@lowering.wraps(stmts.cosh)
def cosh(x: float) -> float: ...


@lowering.wraps(stmts.degrees)
def degrees(x: float) -> float: ...


@lowering.wraps(stmts.erf)
def erf(x: float) -> float: ...


@lowering.wraps(stmts.erfc)
def erfc(x: float) -> float: ...


@lowering.wraps(stmts.exp)
def exp(x: float) -> float: ...


@lowering.wraps(stmts.expm1)
def expm1(x: float) -> float: ...


@lowering.wraps(stmts.fabs)
def fabs(x: float) -> float: ...


@lowering.wraps(stmts.floor)
def floor(x: float) -> float: ...


@lowering.wraps(stmts.fmod)
def fmod(x: float, y: float) -> float: ...


@lowering.wraps(stmts.gamma)
def gamma(x: float) -> float: ...


@lowering.wraps(stmts.isfinite)
def isfinite(x: float) -> float: ...


@lowering.wraps(stmts.isinf)
def isinf(x: float) -> float: ...


@lowering.wraps(stmts.isnan)
def isnan(x: float) -> float: ...


@lowering.wraps(stmts.lgamma)
def lgamma(x: float) -> float: ...


@lowering.wraps(stmts.log10)
def log10(x: float) -> float: ...


@lowering.wraps(stmts.log1p)
def log1p(x: float) -> float: ...


@lowering.wraps(stmts.log2)
def log2(x: float) -> float: ...


@lowering.wraps(stmts.pow)
def pow(x: float, y: float) -> float: ...


@lowering.wraps(stmts.radians)
def radians(x: float) -> float: ...


@lowering.wraps(stmts.remainder)
def remainder(x: float, y: float) -> float: ...


@lowering.wraps(stmts.sin)
def sin(x: float) -> float: ...


@lowering.wraps(stmts.sinh)
def sinh(x: float) -> float: ...


@lowering.wraps(stmts.sqrt)
def sqrt(x: float) -> float: ...


@lowering.wraps(stmts.tan)
def tan(x: float) -> float: ...


@lowering.wraps(stmts.tanh)
def tanh(x: float) -> float: ...


@lowering.wraps(stmts.trunc)
def trunc(x: float) -> float: ...


@lowering.wraps(stmts.ulp)
def ulp(x: float) -> float: ...



from typing import (
    TYPE_CHECKING,
    Any,
    Type,
    Union,
    Generic,
    TypeVar,
    Callable,
    TypeAlias,
    overload,
)
from dataclasses import dataclass

from kirin import types
from kirin.ir import Attribute, Statement
from kirin.interp.value import StatementResult

if TYPE_CHECKING:
    from kirin.interp.base import FrameABC, BaseInterpreter
    from kirin.interp.table import MethodTable

MethodTableSelf = TypeVar("MethodTableSelf", bound="MethodTable")
InterpreterType = TypeVar("InterpreterType", bound="BaseInterpreter")
FrameType = TypeVar("FrameType", bound="FrameABC")
StatementType = TypeVar("StatementType", bound=Statement)
AttributeType = TypeVar("AttributeType", bound=Attribute)
MethodFunction: TypeAlias = Callable[
    [MethodTableSelf, InterpreterType, FrameType, StatementType], StatementResult
]
AttributeFunction: TypeAlias = Callable[
    [MethodTableSelf, InterpreterType, AttributeType], Any
]


@dataclass(frozen=True)
class Signature:
    """Signature of a statement."""

    stmt: Type[Statement]
    args: tuple[types.TypeAttribute, ...] | None = None

    def __repr__(self):
        if self.args:
            return f"{self.stmt.__name__}[{', '.join(map(repr, self.args))}]"
        else:
            return f"{self.stmt.__name__}[...]"


SigType = TypeVar("SigType")
ImplType = TypeVar("ImplType")


@dataclass
class Def(Generic[SigType, ImplType]):
    """Base class of an interpreter implementation definition."""

    signature: SigType
    impl: ImplType


@dataclass
class ImplDef(Def[tuple[Signature, ...], "MethodFunction"]):
    """Definition of an interpreter implementation for a statement."""

    parent: Type[Statement]

    def __repr__(self):
        if self.parent.dialect:
            return f"interp {self.parent.dialect.name}.{self.parent.name}"
        else:
            return f"interp {self.parent.name}"

    def __call__(self, *args, **kwargs):
        return self.impl(*args, **kwargs)


@dataclass
class AttributeImplDef(Def[type[Attribute], "AttributeFunction"]):
    """Definition of an interpreter implementation for an attribute."""

    def __repr__(self):
        if self.signature.dialect:
            return f"attribute impl {self.signature.dialect.name}.{self.signature.name}"
        else:
            return f"attribute impl {self.signature.name}"


StatementType = TypeVar("StatementType", bound=Statement)
HeadType = TypeVar("HeadType")


class impl(Generic[HeadType]):
    """Decorator to define an interpreter implementation for a statement or attribute.

    !!! note
        While the `impl` decorator accepts both statements and attributes, and optionally
        statements with its type signature, unlike a programming language, the actual
        dispatch behavior given an instance of a statement or attribute is defined by the
        implementation of the interpreter
        (via [`lookup_registry`][kirin.interp.base.BaseInterpreter.lookup_registry]).

    # Example

    ```python
    @dialect.register
    class MyMethods(interp.MethodTable):
        @impl(Add)
        def interp_add(
            self,
            interp: Interpreter,
            frame: Frame,
            stmt: Add,
        ) -> StatementResult:
            ...
    ```
    """

    # TODO: validate only concrete types are allowed here

    def __init__(
        self, stmt_or_attribute: Type[HeadType], *args: types.TypeAttribute
    ) -> None:
        if args and issubclass(stmt_or_attribute, Attribute):
            raise ValueError("Attributes do not take arguments")
        self.stmt_or_attribute: type[HeadType] = stmt_or_attribute
        self.args = args

    @overload
    def __call__(
        self,
        func: Union[
            Callable[
                [MethodTableSelf, InterpreterType, FrameType, StatementType],
                StatementResult,
            ],
            ImplDef,
        ],
    ) -> ImplDef: ...

    @overload
    def __call__(
        self,
        func: Union[
            Callable[
                [MethodTableSelf, InterpreterType, AttributeType],
                Any,
            ],
            AttributeImplDef,
        ],
    ) -> AttributeImplDef: ...

    def __call__(
        self,
        func: Union[
            Callable[
                [MethodTableSelf, InterpreterType, FrameType, StatementType],
                StatementResult,
            ],
            Callable[
                [MethodTableSelf, InterpreterType, AttributeType],
                Any,
            ],
            ImplDef,
            AttributeImplDef,
        ],
    ) -> Def:
        if issubclass(self.stmt_or_attribute, Statement):
            return self._impl_statement(self.stmt_or_attribute, func)
        elif issubclass(self.stmt_or_attribute, Attribute):
            return self._impl_attribute(self.stmt_or_attribute, func)
        else:
            raise ValueError(f"Invalid statement type {self.stmt_or_attribute}")

    def _impl_attribute(
        self,
        attr: Type[Attribute],
        func: Union[Callable, Def],
    ) -> AttributeImplDef:
        if isinstance(func, Def):
            return AttributeImplDef(attr, func.impl)
        else:
            return AttributeImplDef(attr, func)

    def _impl_statement(
        self,
        stmt: Type[Statement],
        func: Union[Callable, Def],
    ) -> ImplDef:
        if self.args:
            sig = Signature(stmt, self.args)
        else:
            sig = Signature(stmt)

        if isinstance(func, Def):
            return ImplDef(func.signature + (sig,), func.impl, stmt)
        else:
            return ImplDef((sig,), func, stmt)



from .abc import Result as Result, LoweringABC as LoweringABC
from .frame import Frame as Frame
from .state import State as State
from .exception import BuildError as BuildError
from .python.traits import (
    FromPythonCall as FromPythonCall,
    FromPythonWith as FromPythonWith,
    FromPythonRangeLike as FromPythonRangeLike,
    FromPythonWithSingleItem as FromPythonWithSingleItem,
)
from .python.binding import wraps as wraps
from .python.dialect import FromPythonAST as FromPythonAST, akin as akin
from .python.lowering import Python as Python



from typing import TYPE_CHECKING, Callable, Iterable
from dataclasses import dataclass

from kirin.lowering.python.dialect import FromPythonAST

if TYPE_CHECKING:
    import ast

    from kirin import lowering
    from kirin.ir import Attribute
    from kirin.ir.group import DialectGroup
    from kirin.ir.nodes import Statement
    from kirin.lowering import FromPythonAST
    from kirin.interp.impl import Signature
    from kirin.interp.table import MethodTable
    from kirin.lowering.python.dialect import LoweringTransform


@dataclass
class StatementImpl:
    parent: "MethodTable"
    impl: Callable

    def __call__(self, interp, frame, stmt: "Statement"):
        return self.impl(self.parent, interp, frame, stmt)

    def __repr__(self) -> str:
        return f"method impl `{self.impl.__name__}` in {repr(self.parent.__class__)}"


@dataclass
class AttributeImpl:
    parent: "MethodTable"
    impl: Callable

    def __call__(self, interp, attr: "Attribute"):
        return self.impl(self.parent, interp, attr)

    def __repr__(self) -> str:
        return f"attribute impl `{self.impl.__name__}` in {repr(self.parent.__class__)}"


@dataclass
class CallLoweringFunction:
    parent: "FromPythonAST"
    func: "LoweringTransform"

    def __call__(
        self, state: "lowering.State[ast.AST]", node: "ast.Call"
    ) -> "lowering.Result":
        return self.func(self.parent, state, node)


@dataclass
class LoweringRegistry:
    ast_table: dict[str, FromPythonAST]
    callee_table: dict[object, CallLoweringFunction]


@dataclass
class InterpreterRegistry:
    attributes: dict[type["Attribute"], "AttributeImpl"]
    statements: dict["Signature", "StatementImpl"]


@dataclass
class Registry:
    """Proxy class to build different registries from a dialect group."""

    dialects: "DialectGroup"
    """The dialect group to build the registry from."""

    def ast(self, keys: Iterable[str]) -> LoweringRegistry:
        """select the dialect lowering interpreters for the given key.

        Args:
            keys (Iterable[str]): the keys to search for in the dialects

        Returns:
            a map of dialects to their lowering interpreters
        """
        ret: dict[str, "FromPythonAST"] = {}
        callee_table: dict[object, CallLoweringFunction] = {}
        from_ast = None
        for dialect in self.dialects.data:
            for key in keys:
                if key in dialect.lowering:
                    from_ast = dialect.lowering[key]
                    break

            if from_ast is None:
                msg = ",".join(keys)
                raise KeyError(f"Lowering not found for {msg}")

            for name in from_ast.names:
                if name.startswith("lower_Call_"):
                    continue  # ignore the Call nodes
                if name in ret:
                    raise KeyError(f"Lowering {name} already exists")
                ret[name] = from_ast

            for obj, trans in from_ast.callee_table.items():
                if obj in callee_table:
                    raise KeyError(f"Lowering {obj} already exists")
                callee_table[obj] = CallLoweringFunction(from_ast, trans.func)
        return LoweringRegistry(ret, callee_table)

    def interpreter(self, keys: Iterable[str]):
        """select the dialect interpreter for the given key.

        Args:
            keys (Iterable[str]): the keys to search for in the dialects

        Returns:
            a map of statement signatures to their interpretation functions,
            and a map of dialects to their fallback interpreters.
        """
        attributes: dict[type["Attribute"], "AttributeImpl"] = {}
        table: dict["Signature", "StatementImpl"] = {}
        for dialect in self.dialects.data:
            dialect_table = None
            for key in keys:
                if key not in dialect.interps:
                    continue

                dialect_table = dialect.interps[key]
                for sig, func in dialect_table.attribute.items():
                    if sig not in attributes:
                        attributes[sig] = AttributeImpl(dialect_table, func)

                for sig, func in dialect_table.table.items():
                    if sig not in table:
                        table[sig] = StatementImpl(dialect_table, func)

        return InterpreterRegistry(attributes, table)



from kirin.passes.abc import Pass as Pass
from kirin.passes.fold import Fold as Fold
from kirin.passes.typeinfer import TypeInfer as TypeInfer

from .default import Default as Default
from .hint_const import HintConst as HintConst



"""Module dialect provides a simple module
that is roughly a list of function statements.

This dialect provides the dialect necessary for compiling a function into
lower-level IR with all its callee functions.
"""

from kirin import ir, types, interp
from kirin.decl import info, statement
from kirin.print import Printer
from kirin.analysis import TypeInference

from ._pprint_helper import pprint_calllike

dialect = ir.Dialect("module")


@statement(dialect=dialect)
class Module(ir.Statement):
    traits = frozenset(
        {ir.IsolatedFromAbove(), ir.SymbolTable(), ir.SymbolOpInterface()}
    )
    sym_name: str = info.attribute()
    entry: str = info.attribute()
    body: ir.Region = info.region(multi=False)


@statement(dialect=dialect)
class Invoke(ir.Statement):
    """A special statement that represents
    a function calling functions by symbol name.

    Note:
        This statement is here for completeness, for interpretation,
        it is recommended to rewrite this statement into a `func.Invoke`
        after looking up the symbol table.
    """

    callee: str = info.attribute()
    inputs: tuple[ir.SSAValue, ...] = info.argument()
    kwargs: tuple[str, ...] = info.attribute()
    result: ir.ResultValue = info.result()

    def print_impl(self, printer: Printer) -> None:
        pprint_calllike(self, self.callee, printer)

    def verify(self) -> None:
        if self.kwargs:
            for name in self.kwargs:
                if name not in self.callee:
                    raise ir.ValidationError(
                        self,
                        f"method {self.callee} does not have argument {name}",
                    )
        elif len(self.callee) - 1 != len(self.args):
            raise ir.ValidationError(
                self,
                f"expected {len(self.callee)} arguments, got {len(self.args)}",
            )


@dialect.register
class Concrete(interp.MethodTable):

    @interp.impl(Module)
    def interp_Module(
        self, interp: interp.Interpreter, frame: interp.Frame, stmt: Module
    ):
        for stmt_ in stmt.body.blocks[0].stmts:
            if (trait := stmt.get_trait(ir.SymbolOpInterface)) is not None:
                interp.symbol_table[trait.get_sym_name(stmt_).data] = stmt_
        return ()

    @interp.impl(Invoke)
    def interp_Invoke(
        self, interpreter: interp.Interpreter, frame: interp.Frame, stmt: Invoke
    ):
        callee = interpreter.symbol_table.get(stmt.callee)
        if callee is None:
            raise interp.InterpreterError(f"symbol {stmt.callee} not found")

        trait = callee.get_trait(ir.CallableStmtInterface)
        if trait is None:
            raise interp.InterpreterError(
                f"{stmt.callee} is not callable, got {callee.__class__.__name__}"
            )

        body = trait.get_callable_region(callee)
        mt = ir.Method(
            mod=None,
            py_func=None,
            sym_name=stmt.callee,
            arg_names=[
                arg.name or str(idx) for idx, arg in enumerate(body.blocks[0].args)
            ],
            dialects=interpreter.dialects,
            code=stmt,
        )
        return interpreter.run_method(mt, frame.get_values(stmt.inputs))


@dialect.register(key="typeinfer")
class TypeInfer(interp.MethodTable):

    @interp.impl(Module)
    def typeinfer_Module(
        self, interp: TypeInference, frame: interp.Frame, stmt: Module
    ):
        for stmt_ in stmt.body.blocks[0].stmts:
            if (trait := stmt.get_trait(ir.SymbolOpInterface)) is not None:
                interp.symbol_table[trait.get_sym_name(stmt_).data] = stmt_
        return ()

    @interp.impl(Invoke)
    def typeinfer_Invoke(
        self, interp: TypeInference, frame: interp.Frame, stmt: Invoke
    ):
        callee = interp.symbol_table.get(stmt.callee)
        if callee is None:
            return (types.Bottom,)

        trait = callee.get_trait(ir.CallableStmtInterface)
        if trait is None:
            return (types.Bottom,)

        body = trait.get_callable_region(callee)
        mt = ir.Method(
            mod=None,
            py_func=None,
            sym_name=stmt.callee,
            arg_names=[
                arg.name or str(idx) for idx, arg in enumerate(body.blocks[0].args)
            ],
            dialects=interp.dialects,
            code=stmt,
        )
        interp.run_method(mt, mt.arg_types)
        return tuple(result.type for result in callee.results)



from ... import ir
from .stmts import For, Yield, IfElse
from ...rewrite.abc import RewriteRule, RewriteResult


class ScfToCfRule(RewriteRule):

    def rewrite_ifelse(
        self, node: ir.Region, block_idx: int, curr_block: ir.Block, stmt: IfElse
    ):
        from kirin.dialects import cf

        # create a new block for entering the if statement
        entry_block = ir.Block()
        for arg in curr_block.args:
            arg.replace_by(entry_block.args.append_from(arg.type, arg.name))

        # delete the args of the old block and replace with the result of the # if statement
        for arg in curr_block.args:
            curr_block.args.delete(arg)

        for arg in stmt.results:
            arg.replace_by(curr_block.args.append_from(arg.type, arg.name))

        (then_block := stmt.then_body.blocks[0]).detach()
        (else_block := stmt.else_body.blocks[0]).detach()

        entry_block.stmts.append(
            cf.ConditionalBranch(
                cond=stmt.cond,
                then_arguments=tuple(stmt.args),
                then_successor=then_block,
                else_arguments=tuple(stmt.args),
                else_successor=else_block,
            )
        )

        # insert the then/else blocks and add branch to the current block
        # if the last statement of the then block is a yield
        if isinstance(last_stmt := else_block.last_stmt, Yield):
            last_stmt.replace_by(
                cf.Branch(
                    arguments=tuple(last_stmt.args),
                    successor=curr_block,
                )
            )

        if isinstance(last_stmt := then_block.last_stmt, Yield):
            last_stmt.replace_by(
                cf.Branch(
                    arguments=tuple(last_stmt.args),
                    successor=curr_block,
                )
            )

        node.blocks.insert(block_idx, curr_block)
        node.blocks.insert(block_idx, else_block)
        node.blocks.insert(block_idx, then_block)

        curr_stmt = stmt
        next_stmt = stmt.prev_stmt
        curr_stmt.delete()

        return next_stmt, entry_block

    def rewrite_for(
        self, node: ir.Region, block_idx: int, curr_block: ir.Block, stmt: For
    ):
        from kirin.dialects import cf, py, func

        (body_block := stmt.body.blocks[0]).detach()

        entry_block = ir.Block()
        for arg in curr_block.args:
            arg.replace_by(entry_block.args.append_from(arg.type, arg.name))

        # Get iterator from iterable object
        entry_block.stmts.append(iterable_stmt := py.iterable.Iter(stmt.iterable))
        entry_block.stmts.append(const_none := func.ConstantNone())
        last_stmt = entry_block.last_stmt
        entry_block.stmts.append(
            next_stmt := py.iterable.Next(iterable_stmt.expect_one_result())
        )
        entry_block.stmts.append(
            loop_cmp := py.cmp.Is(next_stmt.expect_one_result(), const_none.result)
        )
        entry_block.stmts.append(
            cf.ConditionalBranch(
                cond=loop_cmp.result,
                then_arguments=tuple(stmt.initializers),
                then_successor=curr_block,
                else_arguments=(next_stmt.expect_one_result(),)
                + tuple(stmt.initializers),
                else_successor=body_block,
            )
        )

        for arg in curr_block.args:
            curr_block.args.delete(arg)

        for arg in stmt.results:
            arg.replace_by(curr_block.args.append_from(arg.type, arg.name))

        if isinstance(last_stmt := body_block.last_stmt, Yield):
            (
                next_stmt := py.iterable.Next(iterable_stmt.expect_one_result())
            ).insert_before(last_stmt)
            (
                loop_cmp := py.cmp.Is(next_stmt.expect_one_result(), const_none.result)
            ).insert_before(last_stmt)
            last_stmt.replace_by(
                cf.ConditionalBranch(
                    cond=loop_cmp.result,
                    else_arguments=(next_stmt.expect_one_result(),)
                    + tuple(last_stmt.args),
                    else_successor=body_block,
                    then_arguments=tuple(last_stmt.args),
                    then_successor=curr_block,
                )
            )

        # insert the body block and add branch to the current block
        node.blocks.insert(block_idx, curr_block)
        node.blocks.insert(block_idx, body_block)

        curr_stmt = stmt
        next_stmt = stmt.prev_stmt
        curr_stmt.delete()

        return next_stmt, entry_block

    def rewrite_ssacfg(self, node: ir.Region):

        has_done_something = False

        for block_idx in range(len(node.blocks)):

            block = node.blocks.pop(block_idx)

            stmt = block.last_stmt
            if stmt is None:
                continue

            curr_block = ir.Block()

            for arg in block.args:
                arg.replace_by(curr_block.args.append_from(arg.type, arg.name))

            while stmt is not None:
                if isinstance(stmt, For):
                    has_done_something = True
                    stmt, curr_block = self.rewrite_for(
                        node, block_idx, curr_block, stmt
                    )

                elif isinstance(stmt, IfElse):
                    has_done_something = True
                    stmt, curr_block = self.rewrite_ifelse(
                        node, block_idx, curr_block, stmt
                    )
                else:
                    curr_stmt = stmt
                    stmt = stmt.prev_stmt
                    curr_stmt.detach()

                    if curr_block.first_stmt is None:
                        curr_block.stmts.append(curr_stmt)
                    else:
                        curr_stmt.insert_before(curr_block.first_stmt)

            # if the last block is empty, remove it
            if curr_block.parent is None and curr_block.first_stmt is not None:
                node.blocks.insert(block_idx, curr_block)

        return RewriteResult(has_done_something=has_done_something)

    def rewrite_Statement(self, node: ir.Statement) -> RewriteResult:
        if isinstance(node, (For, IfElse)):
            # do not do rewrite in scf regions
            return RewriteResult()

        result = RewriteResult()
        for region in node.regions:
            result = result.join(self.rewrite_ssacfg(region))

        return result



from __future__ import annotations


# errors
class InterpreterError(Exception):
    """Generic interpreter error.

    This is the base class for all interpreter errors.
    """

    pass


class FuelExhaustedError(InterpreterError):
    """An error raised when the interpreter runs out of fuel."""

    pass



import math

from kirin.prelude import basic_no_opt
from kirin.dialects import py


def test_math_pi():
    @basic_no_opt
    def main():
        return math.pi

    stmt = main.callable_region.blocks[0].stmts.at(0)
    assert isinstance(stmt, py.Constant)
    assert stmt.value.unwrap() == math.pi



"""Control flow dialect.

This dialect provides a low-level control flow representation.

This dialect does not provide any lowering strategies, to lowering
a Python AST to this dialect, use the `kirin.dialects.lowering.cf` dialect
with this dialect.
"""

from kirin.dialects.cf import (
    emit as emit,
    constprop as constprop,
    typeinfer as typeinfer,
)
from kirin.dialects.cf.stmts import (
    Branch as Branch,
    ConditionalBranch as ConditionalBranch,
)
from kirin.dialects.cf.interp import CfInterpreter as CfInterpreter
from kirin.dialects.cf.dialect import dialect as dialect



import os
import math
import inspect
import textwrap
from pathlib import Path

import black

# NOTE: typeinfer and lowering should be the default, so we don't generate them.


def builtin_math_functions():
    for name, obj in inspect.getmembers(math):
        # skip some special cases for now
        if name in (
            "prod",
            "perm",
            "modf",
            "ldexp",
            "lcm",
            "isqrt",
            "isclose",
            "gcd",
            "fsum",
            "frexp",
            "factorial",
            "acosh",
            "comb",
            "dist",
            "sumprod",
            "nextafter",
            # 3.10 compat
            "cbrt",
            "exp2",
        ):
            continue

        if inspect.isbuiltin(obj):
            try:
                sig = inspect.signature(obj)
                yield name, obj, sig
            except:  # noqa: E722
                continue


with open(os.path.join(os.path.dirname(__file__), "stmts.py"), "w") as f:
    f.write("# This file is generated by gen.py\n")
    f.write("from kirin import ir, types, lowering2\n")
    f.write("from kirin.decl import statement, info\n")
    f.write("from kirin.dialects.math.dialect import dialect\n")
    f.write("\n")
    for name, obj, sig in builtin_math_functions():
        fields = "\n".join(
            [
                f"    {arg} : ir.SSAValue = info.argument(types.Float)"
                for arg in sig.parameters.keys()
            ]
        )
        f.write(
            textwrap.dedent(
                f"""
@statement(dialect=dialect)
class {name}(ir.Statement):
    \"\"\"{name} statement, wrapping the math.{name} function
    \"\"\"
    name = "{name}"
    traits = frozenset({{ir.Pure(), lowering2.FromPythonCall()}})
{fields}
    result: ir.ResultValue = info.result(types.Float)
"""
            )
        )


with open(os.path.join(os.path.dirname(__file__), "interp.py"), "w") as f:
    f.write("# This file is generated by gen.py\n")
    f.write("import math\n")
    f.write("from kirin.dialects.math.dialect import dialect\n")
    f.write("from kirin.dialects.math import stmts\n")
    f.write("from kirin.interp import MethodTable, Frame, impl\n")
    f.write("\n")

    implements = []
    for name, obj, sig in builtin_math_functions():
        fields = ", ".join(
            [f"values[{idx}]" for idx, _ in enumerate(sig.parameters.keys())]
        )
        implements.append(
            f"""
    @impl(stmts.{name})
    def {name}(self, interp, frame: Frame, stmt: stmts.{name}):
        values = frame.get_values(stmt.args)
        return (math.{name}({fields}),)"""
        )

    # Write the interpreter class
    methods = "\n\n".join(implements)
    f.write(
        f"""
@dialect.register
class MathMethodTable(MethodTable):
{methods}
"""
    )

# __init__.py
with open(os.path.join(os.path.dirname(__file__), "__init__.py"), "w") as f:
    f.write('"math dialect, modeling functions in python\'s `math` stdlib"')
    f.write("# This file is generated by gen.py\n")
    f.write("from kirin.dialects.math.dialect import dialect as dialect\n")
    f.write("from . import stmts as stmts, interp as interp\n")
    f.write("import math as pymath\n")
    f.write("pi = pymath.pi\n")
    f.write("e = pymath.e\n")
    f.write("tau = pymath.tau\n")
    f.write("from kirin import lowering2\n")

    for name, obj, sig in builtin_math_functions():
        f.write(
            textwrap.dedent(
                f"""
        @lowering2.wraps(stmts.{name})
        def {name}({", ".join(f"{arg}: float" for arg in sig.parameters.keys())}) -> float: ...
        """
            )
        )
    f.write("\n")

for file in ["__init__.py", "interp.py", "stmts.py"]:
    # format the file in place + using the project config
    black.format_file_in_place(
        Path(os.path.join(os.path.dirname(__file__), file)),
        fast=False,
        mode=black.FileMode(),
    )


# import math as pymath

# from kirin.compile import compile
# from kirin.dialects import math


# # print(math.sin(x=TestValue()))
# # print(inspect.getargspec(math.sin.__init__))
# # print(math.sin.__init__)
# @basic
# def complicated_math_expr(x):
#     return math.sin(math.cos(x) + math.tan(0.5))


# def test_math():
#     complicated_math_expr.code.print()
#     complicated_math_expr.narrow_types()
#     truth = pymath.sin(pymath.cos(1) + pymath.tan(0.5))
#     assert (complicated_math_expr(1) - truth) / truth < 1e-6

# test_basic.py
project_dir = Path(__file__).parent.parent.parent.parent.parent
with open(project_dir.joinpath("test", "dialects", "math", "test_basic.py"), "w") as f:
    f.write("# type: ignore\n")
    f.write("# This file is generated by gen.py\n")
    f.write("import math as pymath\n")
    f.write("from kirin.prelude import basic\n")
    f.write("from kirin.dialects import math\n")
    f.write("\n")
    f.write("\n")

    for name, obj, sig in builtin_math_functions():
        args = ", ".join(arg for arg in sig.parameters.keys())
        inputs = ", ".join("0.42" for _ in sig.parameters.keys())

        f.write(
            textwrap.dedent(
                f"""
                @basic
                def {name}_func({args}):
                    return math.{name}({args})

                def test_{name}():
                    truth = pymath.{name}({inputs})
                    assert ({name}_func({inputs}) - truth) < 1e-6
                """
            )
        )



from kirin import types
from kirin.prelude import basic


@basic
def foo(x: int):
    if x > 1:
        return x + 1
    else:
        return x - 1.0


@basic(typeinfer=True)
def main(x: int):
    return foo(x)


@basic(typeinfer=True)
def moo(x):
    return foo(x)


def test_inter_method_infer():
    assert main.return_type == (types.Int | types.Float)
    # assert moo.arg_types[0] == types.Int  # type gets narrowed based on callee
    assert moo.return_type == (types.Int | types.Float)
    # NOTE: inference of moo should not update foo
    assert foo.arg_types[0] == types.Int
    assert foo.inferred is False
    assert foo.return_type is types.Any



import sys
from abc import ABC, ABCMeta, abstractmethod
from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    TypeVar,
    ClassVar,
    Optional,
    Sequence,
    Generator,
)
from contextlib import contextmanager
from dataclasses import field, dataclass

from typing_extensions import Self, deprecated

from kirin.ir import Block, Method, Region, Statement, DialectGroup, traits
from kirin.exception import KIRIN_INTERP_STATE

from .impl import Signature
from .frame import FrameABC
from .state import InterpreterState
from .value import Successor, ReturnValue, SpecialValue, StatementResult
from .exceptions import InterpreterError

if TYPE_CHECKING:
    from kirin.registry import StatementImpl, InterpreterRegistry

ValueType = TypeVar("ValueType")
FrameType = TypeVar("FrameType", bound=FrameABC)


class InterpreterMeta(ABCMeta):
    """A metaclass for interpreters."""

    pass


@dataclass
class BaseInterpreter(ABC, Generic[FrameType, ValueType], metaclass=InterpreterMeta):
    """A base class for interpreters.

    This class defines the basic structure of an interpreter. It is
    designed to be subclassed to provide the actual implementation of
    the interpreter.

    ### Required Overrides
    When subclassing, if the subclass does not contain `ABC`,
    the subclass must define the following attributes:

    - `keys`: a list of strings that defines the order of dialects to select from.
    - `void`: the value to return when the interpreter evaluates nothing.
    """

    keys: ClassVar[list[str]]
    """The name of the interpreter to select from dialects by order.
    """
    Frame: ClassVar[type[FrameABC]] = field(init=False)
    """The type of the frame to use for this interpreter.
    """
    void: ValueType = field(init=False)
    """What to return when the interpreter evaluates nothing.
    """
    dialects: DialectGroup
    """The dialects to interpret.
    """
    fuel: int | None = field(default=None, kw_only=True)
    """The fuel limit for the interpreter.
    """
    debug: bool = field(default=False, kw_only=True)
    """Whether to enable debug mode.
    """
    max_depth: int = field(default=128, kw_only=True)
    """The maximum depth of the interpreter stack.
    """
    max_python_recursion_depth: int = field(default=8192, kw_only=True)
    """The maximum recursion depth of the Python interpreter.
    """

    # global states
    registry: "InterpreterRegistry" = field(init=False, compare=False)
    """The interpreter registry.
    """
    symbol_table: dict[str, Statement] = field(init=False, compare=False)
    """The symbol table.
    """
    state: InterpreterState[FrameType] = field(init=False, compare=False)
    """The interpreter state.
    """

    # private
    _eval_lock: bool = field(default=False, init=False, compare=False)

    def __post_init__(self) -> None:
        self.registry = self.dialects.registry.interpreter(keys=self.keys)

    def initialize(self) -> Self:
        """Initialize the interpreter global states. This method is called right upon
        calling [`run`][kirin.interp.base.BaseInterpreter.run] to initialize the
        interpreter global states.

        !!! note "Default Implementation"
            This method provides default behavior but may be overridden by subclasses
            to customize or extend functionality.
        """
        self.symbol_table: dict[str, Statement] = {}
        self.state: InterpreterState[FrameType] = InterpreterState()
        return self

    @abstractmethod
    def initialize_frame(
        self, code: Statement, *, has_parent_access: bool = False
    ) -> FrameType:
        """Create a new frame for the given method."""
        ...

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()
        if ABC in cls.__bases__:
            return

        if not hasattr(cls, "keys"):
            raise TypeError(f"keys is not defined for class {cls.__name__}")
        if not hasattr(cls, "void"):
            raise TypeError(f"void is not defined for class {cls.__name__}")

    def run(
        self,
        mt: Method,
        args: tuple[ValueType, ...],
        kwargs: dict[str, ValueType] | None = None,
    ) -> ValueType:
        """Run a method. This is the main entry point of the interpreter.

        Args:
            mt (Method): the method to run.
            args (tuple[ValueType]): the arguments to the method, does not include self.
            kwargs (dict[str, ValueType], optional): the keyword arguments to the method.

        Returns:
            Result[ValueType]: the result of the method.
        """
        if self._eval_lock:
            raise InterpreterError(
                "recursive eval is not allowed, use run_method instead"
            )

        self._eval_lock = True
        self.initialize()
        current_recursion_limit = sys.getrecursionlimit()
        sys.setrecursionlimit(self.max_python_recursion_depth)
        args = self.get_args(mt.arg_names[len(args) + 1 :], args, kwargs)
        try:
            _, results = self.run_method(mt, args)
        except Exception as e:
            # NOTE: insert the interpreter state into the exception
            # so we can print the stack trace
            setattr(e, KIRIN_INTERP_STATE, self.state)
            raise e
        finally:
            self._eval_lock = False
            sys.setrecursionlimit(current_recursion_limit)
        return results

    def run_stmt(
        self, stmt: Statement, args: tuple[ValueType, ...]
    ) -> StatementResult[ValueType]:
        """execute a statement with arguments in a new frame.

        Args:
            stmt (Statement): the statement to run.
            args (tuple[ValueType, ...]): the arguments to the statement.

        Returns:
            StatementResult[ValueType]: the result of the statement.
        """
        with self.new_frame(stmt) as frame:
            frame.set_values(stmt.args, args)
            results = self.eval_stmt(frame, stmt)
            return results

    @abstractmethod
    def run_method(
        self, method: Method, args: tuple[ValueType, ...]
    ) -> tuple[FrameType, ValueType]:
        """How to run a method.

        This is defined by subclasses to describe what's the corresponding
        value of a method during the interpretation. Usually, this method
        just calls [`run_callable`][kirin.interp.base.BaseInterpreter.run_callable].

        Args:
            method (Method): the method to run.
            args (tuple[ValueType, ...]): the arguments to the method, does not include self.

        Returns:
            ValueType: the result of the method.
        """
        ...

    def run_callable(
        self, code: Statement, args: tuple[ValueType, ...]
    ) -> tuple[FrameType, ValueType]:
        """Run a callable statement.

        Args:
            code (Statement): the statement to run.
            args (tuple[ValueType, ...]): the arguments to the statement,
                includes self if the corresponding callable region contains a self argument.

        Returns:
            ValueType: the result of the statement.
        """
        if self.state.depth >= self.max_depth:
            return self.eval_recursion_limit(self.state.current_frame)

        interface = code.get_trait(traits.CallableStmtInterface)
        if interface is None:
            raise InterpreterError(f"statement {code.name} is not callable")

        frame = self.initialize_frame(code)
        self.state.push_frame(frame)
        body = interface.get_callable_region(code)
        if not body.blocks:
            return self.state.pop_frame(), self.void
        results = self.run_callable_region(frame, code, body, args)
        return self.state.pop_frame(), results

    def run_callable_region(
        self,
        frame: FrameType,
        code: Statement,
        region: Region,
        args: tuple[ValueType, ...],
    ) -> ValueType:
        """A hook defines how to run the callable region given
        the interpreter context. Frame should be pushed before calling
        this method and popped after calling this method.

        A callable region is a region that can be called as a function.
        Unlike a general region (or the MLIR convention), it always return a value
        to be compatible with the Python convention.
        """
        results = self.run_ssacfg_region(frame, region, args)
        if isinstance(results, ReturnValue):
            return results.value
        elif not results:  # empty result or None
            return self.void
        raise InterpreterError(
            f"callable region {code.name} does not return `ReturnValue`, got {results}"
        )

    @deprecated("use run_succ instead")
    def run_block(self, frame: FrameType, block: Block) -> SpecialValue[ValueType]:
        """Run a block within the current frame.

        Args:
            frame: the current frame.
            block: the block to run.

        Returns:
            SpecialValue: the result of running the block terminator.
        """
        ...

    def run_succ(self, frame: FrameType, succ: Successor) -> SpecialValue[ValueType]:
        """Run a successor within the current frame.
        Args:
            frame: the current frame.
            succ: the successor to run.

        Returns:
            SpecialValue: the result of running the successor.
        """
        ...

    @contextmanager
    def new_frame(
        self, code: Statement, *, has_parent_access: bool = False
    ) -> Generator[FrameType, Any, None]:
        """Create a new frame for the given method and push it to the state.

        Args:
            code (Statement): the statement to run.

        Keyword Args:
            has_parent_access (bool): whether this frame has access to the
                parent frame entries. Defaults to False.

        This is a context manager that creates a new frame, push and pop
        the frame automatically.
        """
        frame = self.initialize_frame(code, has_parent_access=has_parent_access)
        self.state.push_frame(frame)
        try:
            yield frame
        finally:
            self.state.pop_frame()

    @staticmethod
    def get_args(
        left_arg_names, args: tuple[ValueType, ...], kwargs: dict[str, ValueType] | None
    ) -> tuple[ValueType, ...]:
        if kwargs:
            # NOTE: #self# is not user input so it is not
            # in the args, +1 is for self
            for name in left_arg_names:
                args += (kwargs[name],)
        return args

    @staticmethod
    def permute_values(
        arg_names: Sequence[str],
        values: tuple[ValueType, ...],
        kwarg_names: tuple[str, ...],
    ) -> tuple[ValueType, ...]:
        """Permute the arguments according to the method signature and
        the given keyword arguments, where the keyword argument names
        refer to the last n arguments in the values tuple.

        Args:
            arg_names: the argument names
            values: the values tuple (should not contain method itself)
            kwarg_names: the keyword argument names
        """
        n_total = len(values)
        if kwarg_names:
            kwargs = dict(zip(kwarg_names, values[n_total - len(kwarg_names) :]))
        else:
            kwargs = None

        positionals = values[: n_total - len(kwarg_names)]
        args = BaseInterpreter.get_args(
            arg_names[len(positionals) + 1 :], positionals, kwargs
        )
        return args

    def eval_stmt(
        self, frame: FrameType, stmt: Statement
    ) -> StatementResult[ValueType]:
        """Run a statement within the current frame. This is the entry
        point of running a statement. It will look up the statement implementation
        in the dialect registry, or optionally call a fallback implementation.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            StatementResult: the result of running the statement

        Note:
            Overload this method for the following reasons:
            - to change the source tracking information
            - to take control of how to run a statement
            - to change the implementation lookup behavior that cannot acheive
                by overloading [`lookup_registry`][kirin.interp.base.BaseInterpreter.lookup_registry]

        Example:
            * implement an interpreter that only handles MyStmt:
            ```python
                class MyInterpreter(BaseInterpreter):
                    ...
                    def eval_stmt(self, frame: FrameType, stmt: Statement) -> StatementResult[ValueType]:
                        if isinstance(stmt, MyStmt):
                            return self.run_my_stmt(frame, stmt)
                        else:
                            return ()
            ```

        """
        # TODO: update tracking information
        method = self.lookup_registry(frame, stmt)
        if method is not None:
            results = method(self, frame, stmt)
            if self.debug and not isinstance(results, (tuple, SpecialValue)):
                raise InterpreterError(
                    f"method must return tuple or SpecialResult, got {results}"
                )
            return results
        elif stmt.dialect not in self.dialects:
            # NOTE: we should terminate the interpreter because this is a
            # deveoper error, not a user error.
            name = stmt.dialect.name if stmt.dialect else "None"
            raise ValueError(f"dialect {name} is not supported by {self.dialects}")

        return self.eval_stmt_fallback(frame, stmt)

    @deprecated("use eval_stmt_fallback instead")
    def run_stmt_fallback(
        self, frame: FrameType, stmt: Statement
    ) -> StatementResult[ValueType]:
        return self.eval_stmt_fallback(frame, stmt)

    def eval_stmt_fallback(
        self, frame: FrameType, stmt: Statement
    ) -> StatementResult[ValueType]:
        """The fallback implementation of statements.

        This is called when no implementation is found for the statement.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            StatementResult: the result of running the statement

        Note:
            Overload this method to provide a fallback implementation for statements.
        """
        # NOTE: not using f-string here because 3.10 and 3.11 have
        #  parser bug that doesn't allow f-string in raise statement
        raise InterpreterError(
            "no implementation for stmt "
            + stmt.print_str(end="")
            + " from "
            + str(type(self))
        )

    def eval_recursion_limit(self, frame: FrameType) -> tuple[FrameType, ValueType]:
        """Return the value of recursion exception, e.g in concrete
        interpreter, it will raise an exception if the limit is reached;
        in type inference, it will return a special value.
        """
        raise InterpreterError("maximum recursion depth exceeded")

    def build_signature(self, frame: FrameType, stmt: Statement) -> "Signature":
        """build signature for querying the statement implementation."""
        return Signature(stmt.__class__, tuple(arg.type for arg in stmt.args))

    def lookup_registry(
        self, frame: FrameType, stmt: Statement
    ) -> Optional["StatementImpl[Self, FrameType]"]:
        """Lookup the statement implementation in the registry.

        Args:
            frame: the current frame
            stmt: the statement to run

        Returns:
            Optional[StatementImpl]: the statement implementation if found, None otherwise.
        """
        sig = self.build_signature(frame, stmt)
        if sig in self.registry.statements:
            return self.registry.statements[sig]
        elif (class_sig := Signature(stmt.__class__)) in self.registry.statements:
            return self.registry.statements[class_sig]
        return

    @abstractmethod
    def run_ssacfg_region(
        self, frame: FrameType, region: Region, args: tuple[ValueType, ...]
    ) -> tuple[ValueType, ...] | None | ReturnValue[ValueType]:
        """This implements how to run a region with MLIR SSA CFG convention.

        Args:
            frame: the current frame.
            region: the region to run.
            args: the arguments to the region.

        Returns:
            tuple[ValueType, ...] | SpecialValue[ValueType]: the result of running the region.

        when region returns `tuple[ValueType, ...]`, it means the region terminates normally
        with `YieldValue`. When region returns `ReturnValue`, it means the region terminates
        and needs to pop the frame. Region cannot return `Successor` because reference to
        external region is not allowed.
        """
        ...

    class FuelResult(Enum):
        Stop = 0
        Continue = 1

    def consume_fuel(self) -> FuelResult:
        if self.fuel is None:  # no fuel limit
            return self.FuelResult.Continue

        if self.fuel == 0:
            return self.FuelResult.Stop
        else:
            self.fuel -= 1
            return self.FuelResult.Continue



from .list import List2IList as List2IList
from .const import ConstList2IList as ConstList2IList
from .unroll import Unroll as Unroll
from .hint_len import HintLen as HintLen
from .inline_getitem import InlineGetItem as InlineGetItem



from kirin import ir, types as ktypes
from kirin.prelude import structural
from kirin.analysis import TypeInference


# stmt_at and results_at taken from kirin type inference tests with
# minimal modification
def stmt_at(kernel: ir.Method, block_id: int, stmt_id: int) -> ir.Statement:
    return kernel.code.body.blocks[block_id].stmts.at(stmt_id)  # type: ignore


def results_at(kernel: ir.Method, block_id: int, stmt_id: int):
    return stmt_at(kernel, block_id, stmt_id).results


def test_tuple_type_infer():

    @structural(typeinfer=True)
    def test(x: bool):
        a = [True, False, True]
        return (a[0], x)

    typeinfer = TypeInference(structural)
    frame, _ = typeinfer.run_analysis(test)

    assert [frame.entries[result] for result in results_at(test, 0, 1)] == [
        ktypes.Generic(tuple, ktypes.Bool, ktypes.Bool)
    ]
